"""Splits a LookML string into a sequence of tokens."""

from typing import List, Tuple

import lkml.tokens as tokens
from lkml.keys import CHARACTER_TO_TOKEN, EXPR_BLOCK_KEYS


class Lexer:
    """Splits a LookML string into a sequence of tokens.

    Attributes:
        text: Raw LookML to parse, padded with null character to denote end of stream
        index: Position of lexer in characters as it traverses the text
        tokens: Sequence of tokens that contain the relevant chunks of text
        line_number: Position of lexer in lines as it traverses the text

    """

    def __init__(self, text: str):
        """Initializes the Lexer with a LookML string and sets the index.

        Args:
            text: LookML string to be lexed

        """
        self.text: str = text + "\0"
        self.index: int = 0
        self.tokens: List[tokens.Token] = []
        self.line_number: int = 1

    def peek(self) -> str:
        """Returns the character at the current index of the text being lexed."""
        return self.text[self.index]

    def peek_multiple(self, length: int) -> str:
        """Returns the next n characters from the current index in the text being lexed.

        Args:
            length: The number of characters to return

        """
        return self.text[self.index : self.index + length]

    def advance(self, length: int = 1) -> None:
        """Moves the index forward by n characters.

        Args:
            length: The number of positions forward to move the index.

        """
        self.index += length

    def consume(self) -> str:
        """Returns the current index character and advances the index 1 character."""
        self.advance()
        return self.text[self.index - 1]

    def scan(self) -> Tuple[tokens.Token, ...]:
        """Tokenizes LookML into a sequence of tokens.

        This method skips through the text being lexed until it finds a character that
        indicates the start of a new token. It consumes the relevant characters and adds
        the tokens to a sequence until it reaches the end of the text.
        """
        self.tokens.append(tokens.StreamStartToken(self.line_number))
        while True:
            ch = self.peek()
            if ch == "\0":
                self.tokens.append(CHARACTER_TO_TOKEN[ch](self.line_number))
                break
            elif ch in "\n\t ":
                self.tokens.append(self.scan_whitespace())
            elif ch == "#":
                self.advance()
                self.tokens.append(self.scan_comment())
            elif ch == ";":
                if self.peek_multiple(2) == ";;":
                    self.advance(2)
                    self.tokens.append(CHARACTER_TO_TOKEN[ch](self.line_number))
            elif ch == '"':
                self.advance()
                self.tokens.append(self.scan_quoted_literal())
            elif ch in CHARACTER_TO_TOKEN.keys():
                self.advance()
                self.tokens.append(CHARACTER_TO_TOKEN[ch](self.line_number))
            elif self.check_for_expression_block(self.peek_multiple(25)):
                # TODO: Handle edges here with whitespace and comments
                self.tokens.append(self.scan_literal())
                self.advance()
                self.tokens.append(tokens.ValueToken(self.line_number))
                self.tokens.append(self.scan_expression_block())
            else:
                # TODO: This should actually check for valid literals first
                # and throw an error if it doesn't match
                self.tokens.append(self.scan_literal())

        return tuple(self.tokens)

    @staticmethod
    def check_for_expression_block(string: str) -> bool:
        """Returns True if the input string is an expression block."""
        return any(string.startswith(key + ":") for key in EXPR_BLOCK_KEYS)

    def scan_whitespace(self) -> tokens.WhitespaceToken:
        r"""Returns a token from one or more whitespace characters.

        Example:
            >>> lexer = Lexer("\n\n\t Hello")
            >>> lexer.scan_whitespace()
            WhitespaceToken('\n\n\t ')

        """
        chars = ""
        while self.peek() in "\n\t ":
            if self.peek() == "\n":
                self.line_number += 1
            chars += self.consume()
        return tokens.WhitespaceToken(chars, self.line_number)

    def scan_comment(self) -> tokens.CommentToken:
        r"""Returns a token from a comment.

        The initial pound (#) character is consumed in the scan method, so this
        method only scans for a newline or end of file to indicate the end of the token.

        The pound character is added back to the beginning to the token to emphasize
        the importance of any leading whitespace that follows.

        Example:
            >>> lexer = Lexer(" Disregard this line\n")
            >>> lexer.scan_comment()
            CommentToken(# Disregard this line)

        """
        chars = "#"
        while self.peek() not in "\0\n":
            chars += self.consume()
        return tokens.CommentToken(chars, self.line_number)

    def scan_expression_block(self) -> tokens.ExpressionBlockToken:
        """Returns an token from an expression block string.

        This method strips any trailing whitespace from the expression string, since
        Looker usually adds an extra space before the `;;` terminal.

        Example:
            >>> lexer = Lexer("SELECT * FROM ${TABLE} ;;")
            >>> lexer.scan_expression_block()
            ExpressionBlockToken(SELECT * FROM ${TABLE})

        """
        chars = ""
        while self.peek_multiple(2) != ";;":
            if self.peek() == "\n":
                self.line_number += 1
            chars += self.consume()
        return tokens.ExpressionBlockToken(chars, self.line_number)

    def scan_literal(self) -> tokens.LiteralToken:
        """Returns a token from a literal string.

        Example:
            >>> lexer = Lexer("yes")
            >>> lexer.scan_literal()
            LiteralToken(yes)

        """
        chars = ""
        while self.peek() not in "\0 \n\t:}{,]":
            chars += self.consume()
        return tokens.LiteralToken(chars, self.line_number)

    def scan_quoted_literal(self) -> tokens.QuotedLiteralToken:
        """Returns a token from a quoted literal string.

        The initial double quote character is consumed in the scan method, so this
        method only scans for the trailing quote to indicate the end of the token.

        Example:
            >>> lexer = Lexer('Label"')
            >>> lexer.scan_quoted_literal()
            QuotedLiteralToken(Label)

        """
        chars = ""
        while True:
            ch = self.peek()
            if ch == '"':
                break
            elif ch == "\\":
                chars += self.consume()  # Extra consume to skip the escaped character
            elif ch == "\n":
                self.line_number += 1
            chars += self.consume()
        self.advance()
        return tokens.QuotedLiteralToken(chars, self.line_number)



"""Interface classes between the parse tree and a data structure of primitives.

These classes facilitate parsing and generation to and from simple data structures like
lists and dictionaries, and allow users to parse and generate LookML without needing
to interact with the parse tree.

"""

import logging
from typing import Any, Dict, List, Optional, Sequence, Type, Union, cast

from lkml.keys import (
    EXPR_BLOCK_KEYS,
    KEYS_WITH_NAME_FIELDS,
    PLURAL_KEYS,
    QUOTED_LITERAL_KEYS,
    pluralize,
    singularize,
)
from lkml.tree import (
    BlockNode,
    Comma,
    ContainerNode,
    DocumentNode,
    ExpressionSyntaxToken,
    LeftBracket,
    LeftCurlyBrace,
    ListNode,
    PairNode,
    QuotedSyntaxToken,
    RightBracket,
    RightCurlyBrace,
    SyntaxNode,
    SyntaxToken,
)
from lkml.visitors import Visitor

logger = logging.getLogger(__name__)


def flatten(sequence: list) -> list:
    """Flattens a singly-nested list of lists into a list of items."""
    result = []
    for each in sequence:
        if isinstance(each, list):
            result.extend(each)
        else:
            result.append(each)
    return result


class DictVisitor(Visitor):
    """Creates a primitive representation of the parse tree.

    Traverses the parse tree and transforms each node type into a dict. Each dict is
    combined into one nested dict. Also handles the grouping of fields with plural keys
    like ``dimension`` or ``view`` into lists.

    Attributes:
        depth: Tracks the level of nesting.
    """

    def __init__(self):
        self.depth: int = -1  # Tracks the level of nesting

    def update_tree(self, target: Dict, update: Dict) -> None:
        """Add one dictionary to an existing dictionary, handling certain repeated keys.

        This method is primarily responsible for handling repeated keys in LookML like
        `dimension` or `set`, which can exist more than once in LookML but cannot be
        repeated in a Python dictionary.

        This method checks the list of valid repeated keys and combines the values of
        that key in `target` and/or `update` into a list and assigns a plural key (e.g.
        `dimensions` instead of `dimension`).

        Args:
            target: Existing dictionary of parsed LookML
            update: New dictionary to be added to target

        Raises:
            KeyError: If `update` has more than one key
            KeyError: If the key in `update` already exists and would overwrite existing

        """
        keys = tuple(update.keys())
        if len(keys) > 1:
            raise KeyError("Dictionary to update with cannot have multiple keys.")
        key = keys[0]

        if key in PLURAL_KEYS:
            plural_key = pluralize(key)
            if plural_key in target.keys():
                target[plural_key].append(update[key])
            else:
                target[plural_key] = [update[key]]
        elif key in target.keys():
            if self.depth == 0:
                logger.warning(
                    'Multiple declarations of top-level key "%s" found. '
                    "Using the last-declared value.",
                    key,
                )
                target[key] = update[key]
            else:
                raise KeyError(
                    f'Key "{key}" already exists in tree '
                    "and would overwrite the existing value."
                )
        else:
            target[key] = update[key]

    def visit(self, document: DocumentNode) -> Dict[str, Any]:
        return self.visit_container(document.container)

    def visit_container(self, node: ContainerNode) -> Dict[str, Any]:
        """Creates a dict from a container node by visiting its children."""
        container: Dict[str, Any] = {}
        if len(node.items) > 0:
            self.depth += 1
            for item in node.items:
                self.update_tree(container, item.accept(self))
            self.depth -= 1
        return container

    def visit_block(self, node: BlockNode) -> Dict[str, Dict]:
        """Creates a dict from a block node by visiting its children."""
        container_dict = node.container.accept(self) if node.container else {}
        if node.name is not None:
            container_dict["name"] = node.name.accept(self)
        return {node.type.accept(self): container_dict}

    def visit_list(self, node: ListNode) -> Dict[str, List]:
        """Creates a dict from a list node by visiting its children."""
        return {node.type.accept(self): [item.accept(self) for item in node.items]}

    def visit_pair(self, node: PairNode) -> Dict[str, str]:
        """Creates a dict from pair node by visiting its type and value tokens."""
        return {node.type.accept(self): node.value.accept(self)}

    def visit_token(self, token: SyntaxToken) -> str:
        """Creates a string from a syntax token."""
        return str(token.value)


class DictParser:
    """Parses a Python dictionary into a parse tree.

    Review the grammar specified for the Parser class to understand how LookML
    is represented. The grammar details the differences between blocks, pairs, keys,
    and values.

    Attributes:
        parent_key: The name of the key at the previous level in a LookML block.
        level: The number of indentations appropriate for the current position.
        base_indent: Whitespace representing one tab.
        latest_node: The type of the last node to be parsed.

    """

    def __init__(self):
        self.parent_key: str = None
        self.level: int = 0
        self.base_indent: str = " " * 2
        self.latest_node: Optional[Type[SyntaxNode]] = DocumentNode

    def increase_level(self) -> None:
        """Increases the indent level of the current line by one tab.

        This also resets the latest node, mainly for formatting reasons.

        """
        self.latest_node = None
        self.level += 1

    def decrease_level(self) -> None:
        """Decreases the indent level of the current line by one tab."""
        self.level -= 1

    @property
    def indent(self) -> str:
        """Returns the level-adjusted indent."""
        if self.level > 0:
            return self.base_indent * self.level
        else:
            return ""

    @property
    def newline_indent(self) -> str:
        """Returns a newline plus the current indent."""
        return "\n" + self.indent

    @property
    def prefix(self) -> str:
        """Returns the currently appropriate, preceding whitespace."""
        if self.latest_node == DocumentNode:
            return ""
        elif self.latest_node is None:
            return self.newline_indent
        elif self.latest_node == BlockNode:
            return "\n" + self.newline_indent
        else:
            return self.newline_indent

    def is_plural_key(self, key: str) -> bool:
        """Returns True if the key is a repeatable key.

        For example, `dimension` can be repeated, but `sql` cannot be.

        The key `allowed_value` is a special case and changes behavior depending on its
        parent key. If its parent key is `access_grant`, it is a list and cannot be
        repeated. Otherwise, it can be repeated.

        The parent key `query` is also a special case, where children are kept as lists.
        See issue #53.

        Args:
            key: The name of the key to test.

        """
        singular_key = singularize(key)
        return (
            singular_key in PLURAL_KEYS
            and not (
                singular_key == "allowed_value"
                and self.parent_key.rstrip("s") == "access_grant"
            )
            and not (self.parent_key == "query")
        )

    def resolve_filters(self, values: List[dict]) -> Union[List[BlockNode], ListNode]:
        """Parse the key ``filters`` according to the context.

        In LookML, the ``filters`` key is wildly inconsistent and can have three
        different syntaxes. This method determines the syntax that should be used based
        on the context and parses the appropriate node.

        Args:
            values: The contents of the ``filters`` block. Provides context to resolve.

        Returns:
            A block or list node depending on the resolution.

        """
        if "name" in values[0]:
            # This is one or more filter-only field(s), e.g.
            # filter: order_region { type: string }
            blocks = []
            for value in values:
                name = value.pop("name")
                block = self.parse_block(key="filter", items=value, name=name)
                blocks.append(block)
            return blocks
        elif "field" in values[0] and "value" in values[0]:
            # This is the legacy filter syntax, e.g.
            # filters: { field: dimension_name, value: "filter expression" }
            return [self.parse_block(key="filters", items=value) for value in values]
        else:
            # This is the new filter syntax, e.g.
            # filters: [ dimension_name: "filter expression", ... ]
            return self.parse_list(key="filters", values=values)

    def parse(self, obj: Dict[str, Any]) -> DocumentNode:
        """Parses a primitive representation of LookML into a parse tree."""
        nodes = [self.parse_any(key, value) for key, value in obj.items()]
        container = ContainerNode(items=tuple(flatten(nodes)))
        return DocumentNode(container)

    def expand_list(
        self, key: str, values: Sequence
    ) -> List[Union[BlockNode, ListNode, PairNode]]:
        """Expands and parses a list of values for a repeatable key.

        Args:
            key: A repeatable LookML field type (e.g. "views" or "dimension_groups")
            values: A sequence of objects to be parsed

        Returns:
            A list of block, list, or pair nodes, depending on the list's contents.

        """
        # A dictionary with a key "filters" can correspond to multiple syntaxes, so
        # must be handled in a context-aware manner
        if key == "filters":
            values = cast(List[dict], values)
            return flatten([self.resolve_filters(values)])
        else:
            singular_key = singularize(key)
            return flatten([self.parse_any(singular_key, value) for value in values])

    def parse_any(
        self, key: str, value: Union[str, list, tuple, dict]
    ) -> Union[
        List[Union[BlockNode, ListNode, PairNode]], BlockNode, ListNode, PairNode
    ]:
        """Dynamically serializes a Python object based on its type.

        Args:
            key: A LookML field type (e.g. "suggestions" or "hidden")
            value: A string, tuple, or list to serialize

        Raises:
            TypeError: If input value is not of a valid type

        Returns:
            A generator of serialized string chunks

        """
        if isinstance(value, str):
            return self.parse_pair(key, value)
        elif isinstance(value, (list, tuple)):
            if self.is_plural_key(key):
                return self.expand_list(key, value)
            else:
                return self.parse_list(key, value)
        elif isinstance(value, dict):
            if key in KEYS_WITH_NAME_FIELDS or "name" not in value.keys():
                name = None
            else:
                name = value.pop("name")
            return self.parse_block(key, value, name)
        else:
            raise TypeError("Value must be a string, list, tuple, or dict.")

    def parse_block(
        self, key: str, items: Dict[str, Any], name: Optional[str] = None
    ) -> BlockNode:
        """Serializes a dictionary to a LookML block.

        Args:
            key: A LookML field type (e.g. "dimension")
            fields: A dictionary to serialize (e.g. {"sql": "${TABLE}.order_id"})
            name: An optional name of the block (e.g. "order_id")

        Returns:
            A generator of serialized string chunks

        """
        prev_parent_key = self.parent_key
        self.parent_key = key
        latest_node_at_this_level = self.latest_node
        self.increase_level()
        nodes = [self.parse_any(key, value) for key, value in items.items()]
        self.decrease_level()
        self.latest_node = latest_node_at_this_level
        self.parent_key = prev_parent_key

        container = ContainerNode(items=tuple(flatten(nodes)))

        if self.latest_node and self.latest_node != DocumentNode:
            prefix = "\n" + self.newline_indent
        else:
            prefix = self.prefix

        node = BlockNode(
            type=SyntaxToken(key, prefix=prefix),
            left_brace=LeftCurlyBrace(prefix=" " if name else ""),
            right_brace=RightCurlyBrace(
                prefix=self.newline_indent if container.items else ""
            ),
            name=SyntaxToken(name) if name else None,
            container=container,
        )
        self.latest_node = BlockNode
        return node

    def parse_list(self, key: str, values: Sequence[Union[str, Dict]]) -> ListNode:
        """Serializes a sequence to a LookML block.

        Args:
            key: A LookML field type (e.g. "fields")
            values: A sequence to serialize (e.g. ["orders.order_id", "orders.item"])

        Returns:
            A generator of serialized string chunks

        """
        # `suggestions` is only quoted when it's a list, so override the default
        force_quote = True if key == "suggestions" else False
        prev_parent_key = self.parent_key
        self.parent_key = key

        type_token = SyntaxToken(key, prefix=self.prefix)
        right_bracket = RightBracket()
        items: list = []
        pair_mode = False

        # Check the first element to see if it's a single value or a pair
        if values and not isinstance(values[0], (str, int)):
            pair_mode = True

        # Coerce type depending on pair mode value
        if pair_mode:
            items = cast(List[PairNode], items)
        else:
            items = cast(List[SyntaxToken], items)

        # Choose newline delimiting or space delimiting based on contents
        if len(values) >= 5 or pair_mode:
            trailing_comma: Optional[Comma] = Comma()
            self.increase_level()
            for value in values:
                if pair_mode:
                    value = cast(dict, value)
                    # Extract key and value from dictionary with only one key
                    [(key, val)] = value.items()
                    pair: PairNode = self.parse_pair(key, val)
                    items.append(pair)
                else:
                    value = cast(str, value)
                    token: SyntaxToken = self.parse_token(
                        key, value, force_quote, prefix=self.newline_indent
                    )
                    items.append(token)
            self.decrease_level()
            right_bracket = RightBracket(prefix=self.newline_indent)
        else:
            trailing_comma = None
            for i, value in enumerate(values):
                value = cast(str, value)
                if i == 0:
                    token = self.parse_token(key, value, force_quote)
                else:
                    token = self.parse_token(key, value, force_quote, prefix=" ")
                items.append(token)

        self.parent_key = prev_parent_key

        node = ListNode(
            type=type_token,
            left_bracket=LeftBracket(),
            items=tuple(items),
            right_bracket=right_bracket,
            trailing_comma=trailing_comma,
        )
        self.latest_node = ListNode
        return node

    def parse_pair(self, key: str, value: str) -> PairNode:
        """Serializes a key and value to a LookML pair.

        Args:
            key: A LookML field type (e.g. "hidden")
            value: The value string (e.g. "yes")

        Returns:
            A generator of serialized string chunks

        """
        force_quote = True if self.parent_key == "filters" and key != "field" else False
        value_syntax_token: SyntaxToken = self.parse_token(key, value, force_quote)
        node = PairNode(
            type=SyntaxToken(key, prefix=self.prefix), value=value_syntax_token
        )
        self.latest_node = PairNode
        return node

    @staticmethod
    def parse_token(
        key: str,
        value: str,
        force_quote: bool = False,
        prefix: str = "",
        suffix: str = "",
    ) -> SyntaxToken:
        """Parses a value into a token, quoting it if required by the key or forced.

        Args:
            key: A LookML field type (e.g. "hidden")
            value: The value string (e.g. "yes")
            force_quote: True if value should always be quoted

        Returns:
            A generator of serialized string chunks

        """
        if force_quote or key in QUOTED_LITERAL_KEYS:
            return QuotedSyntaxToken(value, prefix=prefix, suffix=suffix)
        elif key in EXPR_BLOCK_KEYS:
            return ExpressionSyntaxToken(value.strip(), prefix=prefix, suffix=suffix)
        else:
            return SyntaxToken(value, prefix=prefix, suffix=suffix)



"""Node and token classes that make up the parse tree."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Counter, Optional, Tuple, Union, cast

from lkml.keys import PLURAL_KEYS


def items_to_str(*items: Any) -> str:
    """Converts each item to a string and joins them together."""
    return "".join(str(item) for item in items)


@dataclass(frozen=True)
class SyntaxToken:
    """Stores a text value with optional prefix or suffix trivia.

    For example, a syntax token might represent meaningful punctuation like a curly
    brace or the type or value of a LookML field. A syntax token can also store trivia,
    comments or whitespace that precede or follow the token value. The parser attempts
    to assign these prefixes and suffixes intelligently to the corresponding tokens.

    Attributes:
        value: The text represented by the token.
        prefix: Comments or whitespace preceding the token.
        suffix: Comments or whitespace following the token.

    """

    value: str
    line_number: Optional[int] = None
    prefix: str = ""
    suffix: str = ""

    def format_value(self) -> str:
        """Returns the value itself, subclasses may modify the value first."""
        return self.value

    def accept(self, visitor: Visitor) -> Any:
        """Accepts a visitor and calls the visitor's token method on itself."""
        return visitor.visit_token(self)

    def __str__(self) -> str:
        return items_to_str(self.prefix, self.format_value(), self.suffix)


@dataclass(frozen=True)
class LeftCurlyBrace(SyntaxToken):
    value: str = "{"


@dataclass(frozen=True)
class RightCurlyBrace(SyntaxToken):
    value: str = "}"


@dataclass(frozen=True)
class Colon(SyntaxToken):
    value: str = ":"


@dataclass(frozen=True)
class LeftBracket(SyntaxToken):
    value: str = "["


@dataclass(frozen=True)
class RightBracket(SyntaxToken):
    value: str = "]"


@dataclass(frozen=True)
class DoubleSemicolon(SyntaxToken):
    value: str = ";;"


@dataclass(frozen=True)
class Comma(SyntaxToken):
    value: str = ","


class QuotedSyntaxToken(SyntaxToken):
    def format_value(self) -> str:
        # Escape double quotes since the whole value is quoted
        return '"' + self.value.replace(r"\"", '"').replace('"', r"\"") + '"'


@dataclass(frozen=True)
class ExpressionSyntaxToken(SyntaxToken):
    prefix: str = " "
    expr_suffix: str = " "

    def __str__(self) -> str:
        return items_to_str(
            self.prefix, self.format_value(), self.expr_suffix, ";;", self.suffix
        )


class SyntaxNode(ABC):
    """Abstract base class for members of the parse tree that have child nodes."""

    @property
    @abstractmethod
    def children(self) -> Optional[Tuple[SyntaxNode, ...]]:
        """Returns all child SyntaxNodes, but not SyntaxTokens."""
        ...

    @property
    @abstractmethod
    def line_number(self) -> Optional[int]:
        """Returns the line number of the first SyntaxToken in the node"""
        ...

    @abstractmethod
    def accept(self, visitor: Visitor) -> Any:
        """Accepts a Visitor that can interact with the node.

        The visitor pattern allows for flexible algorithms that can traverse the tree
        without needing to be defined as methods on the tree itself.

        """
        ...


@dataclass(frozen=True)
class PairNode(SyntaxNode):
    """A simple LookML field, e.g. ``hidden: yes``.

    Attributes:
        type: The field type, the value that precedes the colon.
        value: The field value, the value that follows the colon.
        colon: An optional Colon SyntaxToken. If not supplied, a default colon is
            created with a single space suffix after the colon.

    """

    type: SyntaxToken
    value: SyntaxToken
    colon: Colon = Colon(suffix=" ")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(type='{self.type.value}', value='{self.value.value}')"
        )

    @property
    def children(self) -> None:
        return None

    @property
    def line_number(self) -> Optional[int]:
        return self.type.line_number

    def accept(self, visitor: Visitor) -> Any:
        """Accepts a visitor and calls the visitor's pair method on itself."""
        return visitor.visit_pair(self)

    def __str__(self) -> str:
        return items_to_str(self.type, self.colon, self.value)


@dataclass(frozen=True)
class ListNode(SyntaxNode):
    """A LookML list, enclosed in square brackets. Like ``fields`` or ``filters``.

    Attributes:
        type: The field type, the value that precedes the colon.
        items: A tuple of pair nodes or syntax tokens, depending on the list style.
        left_bracket: A syntax token for the opening bracket "[".
        right_bracket: A syntax token for the closing bracket "]".
        colon: An optional Colon SyntaxToken. If not supplied, a default colon is
            created with a single space suffix after the colon.
        trailing_comma: Include a trailing comma after the last item.

    """

    type: SyntaxToken
    items: Union[Tuple[PairNode, ...], Tuple[SyntaxToken, ...]]
    left_bracket: LeftBracket
    right_bracket: RightBracket
    colon: Colon = Colon(suffix=" ")
    leading_comma: Optional[Comma] = None
    trailing_comma: Optional[Comma] = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(type='{self.type.value}')"

    @property
    def children(self,) -> Tuple[PairNode, ...]:
        if self.items and isinstance(self.items[0], PairNode):
            # Assume that all elements are pairs
            self.items = cast(Tuple[PairNode, ...], self.items)  # type: ignore
            return self.items
        else:
            return tuple()

    @property
    def line_number(self) -> Optional[int]:
        return self.type.line_number

    def accept(self, visitor: Visitor) -> Any:
        """Accepts a visitor and calls the visitor's list method on itself."""
        return visitor.visit_list(self)

    def __str__(self) -> str:
        return items_to_str(
            self.type,
            self.colon,
            self.left_bracket,
            self.leading_comma if self.leading_comma and len(self.items) > 0 else "",
            ",".join(str(item) for item in self.items),
            self.trailing_comma if self.trailing_comma and len(self.items) > 0 else "",
            self.right_bracket,
        )


@dataclass(frozen=True)
class BlockNode(SyntaxNode):
    """A LookML block, enclosed in curly braces. Like ``view`` or ``dimension``.

    Attributes:
        type: The field type, the value that precedes the colon.
        left_brace: A syntax token for the opening brace "{".
        right_brace: A syntax token for the closing brace "}".
        colon: An optional Colon SyntaxToken. If not supplied, a default colon is
            created with a single space suffix after the colon.
        name: An optional name token, the value that follows the colon.
        container: A container node that holds the block's child nodes.

    """

    type: SyntaxToken
    left_brace: LeftCurlyBrace
    right_brace: RightCurlyBrace
    colon: Optional[Colon] = Colon(suffix=" ")
    name: Optional[SyntaxToken] = None
    container: Optional[ContainerNode] = None

    def __repr__(self) -> str:
        name = f"name='{self.name.value}'" if self.name else None
        return f"{self.__class__.__name__}(type='{self.type.value}', {name})"

    @property
    def children(self) -> Tuple[ContainerNode, ...]:
        return (self.container,) if self.container else tuple()

    @property
    def line_number(self) -> Optional[int]:
        return self.type.line_number

    def accept(self, visitor: Visitor) -> Any:
        """Accepts a visitor and calls the visitor's block method on itself."""
        return visitor.visit_block(self)

    def __str__(self) -> str:
        name = self.name or ""
        container = self.container or ""
        return items_to_str(
            self.type, self.colon, name, self.left_brace, container, self.right_brace
        )


@dataclass(frozen=True)
class DocumentNode(SyntaxNode):
    """The root node of the parse tree.

    Attributes:
        container: The top-level container node.
        prefix: Leading whitespace or comments before the document.
        suffix: Trailing whitespace or comments after the document.

    """

    container: ContainerNode
    prefix: str = ""
    suffix: str = ""

    @property
    def children(self) -> Tuple[ContainerNode]:
        return (self.container,)

    @property
    def line_number(self) -> Optional[int]:
        return 1  # Document always starts on the first line

    def accept(self, visitor: Visitor) -> Any:
        """Accepts a visitor and calls the visitor's visit method on itself."""
        return visitor.visit(self)

    def __str__(self) -> str:
        return items_to_str(self.prefix, self.container, self.suffix)


@dataclass(frozen=True)
class ContainerNode(SyntaxNode):
    """A sequence of nodes, either at the top level of a document, or within a block.

    Attributes:
        items: A tuple of the contained nodes.
        top_level: If the container is the top level of the LookML document.

    Raises:
        KeyError: If a key already exists in the tree and would be overwritten.

    """

    items: Tuple[Union[BlockNode, PairNode, ListNode], ...]
    top_level: bool = False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __post_init__(self):
        counter = Counter(item.type.value for item in self.items)
        for key, count in counter.items():
            if not self.top_level and count > 1 and key not in PLURAL_KEYS:
                raise KeyError(
                    f'Key "{key}" already exists in tree and would overwrite the '
                    "existing value."
                )

    @property
    def children(self) -> Tuple[Union[BlockNode, PairNode, ListNode], ...]:
        return self.items

    def line_number(self) -> Optional[int]:
        try:
            first_item = self.items[0]
        except IndexError:
            return None
        else:
            return first_item.line_number

    def accept(self, visitor: Visitor) -> Any:
        """Accepts a visitor and calls the visitor's container method on itself."""
        return visitor.visit_container(self)

    def __str__(self) -> str:
        # TODO: This produces unparseable LookML for unquoted pair values if no suffix
        # For example, hidden: yes + dimension: ... with no whitespace in between
        return items_to_str(*self.items)


class Visitor(ABC):
    """Abstract base class for visitors that interact with the parse tree."""

    @abstractmethod
    def visit(self, document: DocumentNode) -> Any:
        ...

    @abstractmethod
    def visit_container(self, node: ContainerNode) -> Any:
        ...

    @abstractmethod
    def visit_block(self, node: BlockNode) -> Any:
        ...

    @abstractmethod
    def visit_list(self, node: ListNode) -> Any:
        ...

    @abstractmethod
    def visit_pair(self, node: PairNode) -> Any:
        ...

    @abstractmethod
    def visit_token(self, token: SyntaxToken) -> Any:
        ...



"""Parses a sequence of tokenized LookML into a parse tree."""

import logging
from dataclasses import dataclass, field
from functools import wraps
from typing import List, Optional, Sequence, Tuple, Type, Union

import lkml.tokens as tokens
import lkml.tree as tree
import lkml.utils as utils

# Delimiter character used during logging to show the depth of nesting
DELIMITER = ". "
Syntax = Union[tree.SyntaxNode, tree.SyntaxToken]

logger = logging.getLogger(__name__)


def backtrack_if_none(fn):
    """Decorates parsing methods to backtrack to a previous position on failure.

    This method sets a marker at the current position before attempting to run a
    parsing method. If the parsing method fails and returns `None`, it resets the
    index to the marker.

    It also keeps track of the farthest index of progress in case all parsing
    methods fail and we need to return a SyntaxError to the user with a character
    number.

    Args:
        fn (Callable): The method to be decorated for backtracking.

    """

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        mark = self.index
        self.depth += 1
        result = fn(self, *args, **kwargs)
        self.depth -= 1
        if result is None:
            self.progress = self.index if self.index > self.progress else self.progress
            self.jump_to_index(mark)
        return result

    return wrapper


@dataclass
class CommaSeparatedValues:
    """Helper class to store a series of values and a flag for a trailing comma."""

    _values: list = field(default_factory=list)
    trailing_comma: Optional[tree.Comma] = None
    leading_comma: Optional[tree.Comma] = None

    def append(self, value):
        """Add a value to the private _values list."""
        self._values.append(value)

    @property
    def values(self) -> tuple:
        """Return the private _values list, cast to a tuple."""
        return tuple(self._values)


class Parser:
    r"""Parses a sequence of tokenized LookML into a parse tree.

    This parser is a recursive descent parser which uses the grammar listed below (in
    PEG format). Each grammar rule aligns with a corresponding method (e.g.
    parse_expression).

    Grammar:
        * ``expression`` ← ``(block / pair / list)*``
        * ``block`` ← ``key literal? "{" expression "}"``
        * ``pair`` ← ``key value``
        * ``list`` ← ``key "[" csv? "]"``
        * ``csv`` ← ``(literal / quoted_literal) ("," (literal / quoted_literal))* ","?``
        * ``value`` ← ``literal / quoted_literal / expression_block``
        * ``key`` ← ``literal ":"``
        * ``expression_block`` ← ``[^;]* ";;"``
        * ``quoted_literal`` ← ``'"' [^\"]+ '"'``
        * ``literal`` ← ``[0-9A-Za-z_]+``

    Attributes:
        tokens: A sequence of tokens to be parsed.
        index: The position in the token sequence being parsed.
        progress: The farthest index of progress during parsing.
        depth: The level of recursion into nested expressions.
        log_debug: A flag indicating that debug messages should be logged. This flag
            exits to turn off logging flow entirely, which provides a small
            performance gain compared to parsing at a non-debug logging level.

    """

    def __init__(self, stream: Sequence[tokens.Token]):
        """Initializes the Parser with a stream of tokens and sets the index.

        Args:
            stream: Lexed sequence of tokens to be parsed

        Raises:
            TypeError: If an object in the stream is not a valid token

        """
        for token in stream:
            if not isinstance(token, tokens.Token):
                raise TypeError(
                    f"Type {type(token)} for {token} is not a valid token type."
                )
        self.tokens = stream
        self.index: int = 0
        self.progress: int = 0
        self.depth: int = -1
        self.log_debug: bool = logger.isEnabledFor(logging.DEBUG)

    def jump_to_index(self, index: int):
        """Sets the parser index to a specified value."""
        self.index = index

    def peek(self) -> tokens.Token:
        """Returns the token at the current index."""
        return self.tokens[self.index]

    def advance(self, length: int = 1):
        """Moves the index forward by n characters.

        Args:
            length: The number of positions forward to move the index.

        """
        self.index += length

    def consume(self) -> tokens.Token:
        """Returns the current index character and advances the index by 1 token."""
        self.advance()
        return self.tokens[self.index - 1]

    def consume_token_value(self) -> str:
        """Returns the value of the current index token, advancing the index 1 token."""
        token = self.consume()
        if token.value is None:
            raise ValueError("Token %s does not have a consumable value." % token)
        return token.value

    def consume_trivia(self) -> str:
        """Returns all continuous trivia values."""
        trivia = ""
        while True:
            if self.check(tokens.CommentToken, tokens.WhitespaceToken):
                trivia += self.consume_token_value()
            else:
                break
        return trivia

    def check(
        self, *token_types: Type[tokens.Token], skip_trivia: bool = False
    ) -> bool:
        """Compares the current index token type to specified token types.

        Args:
            *token_types: A variable number of token types to check against.
            skip_trivia: Ignore trivia tokens when searching for a match.

        Raises:
            TypeError: If one or more of the token_types are not actual token types

        Returns:
            True if the current token matches one of the token_types.

        """
        # Set a bookmark to skip over trivia tokens before the check if asked
        mark = self.index
        if skip_trivia:
            _ = self.consume_trivia()

        if self.log_debug:
            logger.debug(
                "%sCheck %s == %s",
                (1 + self.depth) * DELIMITER,
                self.peek(),
                " or ".join(t.__name__ for t in token_types),
            )
        for token_type in token_types:
            if not issubclass(token_type, tokens.Token):
                raise TypeError(f"{token_type} is not a valid token type.")

        try:
            token = self.peek()
        except IndexError:
            # Reached the end of the stream
            result = False
        else:
            if type(token) in token_types:
                result = True
            else:
                result = False
        finally:
            if skip_trivia:
                self.jump_to_index(mark)
        return result

    def parse(self) -> tree.DocumentNode:
        """Main method of this class and a wrapper for the container parser.

        Returns:
            A document node, the root node of the LookML parse tree.

        """
        if self.check(tokens.StreamStartToken):
            self.advance()
        prefix = self.consume_trivia()
        container = self.parse_container()
        suffix = self.consume_trivia()
        return tree.DocumentNode(container, prefix, suffix)

    @backtrack_if_none
    def parse_container(self) -> tree.ContainerNode:
        """Returns a container node that contains any number of children.

        Grammar: ``expression`` ← ``(block / pair / list)*``

        Returns:
            A node with the parsed container or None if the grammar doesn't match.

        Raises:
            SyntaxError: If unable to find a matching grammar rule for the stream

        """
        if self.log_debug:
            grammar = "[expression] = (block / pair / list)*"
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)
        items: List[Union[tree.BlockNode, tree.PairNode, tree.ListNode]] = []
        while not self.check(
            tokens.StreamEndToken, tokens.BlockEndToken, skip_trivia=True
        ):
            block = self.parse_block()
            if block is not None:
                items.append(block)
                continue

            pair = self.parse_pair()
            if pair is not None:
                items.append(pair)
                continue

            list_ = self.parse_list()
            if list_ is not None:
                items.append(list_)
                continue

            token = self.tokens[self.progress]
            raise SyntaxError(
                f"Unable to find a matching expression for '{token.id}' "
                f"on line {token.line_number}"
            )

        container = tree.ContainerNode(items=tuple(items), top_level=self.depth == 0)
        if self.log_debug:
            logger.debug("%sSuccessfully parsed expression.", self.depth * DELIMITER)
        return container

    @backtrack_if_none
    def parse_block(self) -> Optional[tree.BlockNode]:
        """Returns a node that represents a LookML block.

        Grammar: ``block`` ← ``key literal? "{" expression "}"``

        Returns:
            A node with the parsed block or None if the grammar doesn't match.

        """
        if self.log_debug:
            grammar = "[block] = key literal? '{' expression '}'"
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        key = self.parse_key()
        if key is None:
            return key

        if self.check(tokens.LiteralToken):
            token = self.consume()
            name: Optional[tree.SyntaxToken] = tree.SyntaxToken(
                token.value, token.line_number
            )
        else:
            name = None

        prefix = self.consume_trivia()
        if self.check(tokens.BlockStartToken):
            self.advance()
            suffix = self.consume_trivia()
            left_brace = tree.LeftCurlyBrace(prefix=prefix, suffix=suffix)
        else:
            return None

        container = self.parse_container()

        prefix = self.consume_trivia()
        if self.check(tokens.BlockEndToken):
            self.advance()
            suffix = self.consume_trivia()
            right_brace = tree.RightCurlyBrace(prefix=prefix, suffix=suffix)

            block = tree.BlockNode(
                type=key[0],
                colon=key[1],
                name=name,
                left_brace=left_brace,
                container=container,
                right_brace=right_brace,
            )

            if self.log_debug:
                logger.debug("%sSuccessfully parsed block.", self.depth * DELIMITER)
            return block
        else:
            return None

    @backtrack_if_none
    def parse_pair(self) -> Optional[tree.PairNode]:
        """Returns a dictionary that represents a LookML key/value pair.

        Grammar: ``pair`` ← ``key value``

        Returns:
            A dictionary with the parsed pair or None if the grammar doesn't match.

        """
        if self.log_debug:
            grammar = "[pair] = key value"
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        key = self.parse_key()
        if key is None:
            return None
        value = self.parse_value()
        if value is None:
            return None

        pair = tree.PairNode(type=key[0], colon=key[1], value=value)
        logger.debug(self.depth * DELIMITER + "Successfully parsed pair.")
        return pair

    @backtrack_if_none
    def parse_key(self) -> Optional[Tuple[tree.SyntaxToken, tree.Colon]]:
        """Returns a syntax token that represents a literal key and colon character.

        Grammar: ``key`` ← ``literal ":"``

        Returns:
            A tuple of syntax tokens with the parsed key and colon or None if the
            grammar doesn't match.

        """
        if self.log_debug:
            grammar = "[key] = literal ':'"
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)
        prefix = self.consume_trivia()
        if self.check(tokens.LiteralToken):
            token = self.consume()
            key = tree.SyntaxToken(token.value, token.line_number, prefix=prefix)
        else:
            return None

        prefix = self.consume_trivia()
        if self.check(tokens.ValueToken):
            token = self.consume()
            suffix = self.consume_trivia()
            colon = tree.Colon(
                line_number=token.line_number, prefix=prefix, suffix=suffix
            )
        else:
            return None

        if self.log_debug:
            logger.debug("%sSuccessfully parsed key.", self.depth * DELIMITER)
        return key, colon

    @backtrack_if_none
    def parse_value(
        self, parse_prefix: bool = False, parse_suffix: bool = False
    ) -> Optional[tree.SyntaxToken]:
        """Returns a syntax token that represents a value.

        Grammar: ``value`` ← ``literal / quoted_literal / expression_block``

        Returns:
            A syntax token with the parsed value or None if the grammar doesn't match.

        """
        if self.log_debug:
            grammar = "[value] = literal / quoted_literal / expression_block"
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        prefix = self.consume_trivia() if parse_prefix else ""

        if self.check(tokens.LiteralToken):
            token = self.consume()
            suffix = self.consume_trivia() if parse_suffix else ""
            if self.log_debug:
                logger.debug("%sSuccessfully parsed value.", self.depth * DELIMITER)
            return tree.SyntaxToken(token.value, token.line_number, prefix, suffix)
        elif self.check(tokens.QuotedLiteralToken):
            token = self.consume()
            suffix = self.consume_trivia() if parse_suffix else ""
            if self.log_debug:
                logger.debug("%sSuccessfully parsed value.", self.depth * DELIMITER)
            return tree.QuotedSyntaxToken(
                token.value, token.line_number, prefix, suffix
            )
        elif self.check(tokens.ExpressionBlockToken):
            token = self.consume()
            expr_prefix, value, expr_suffix = utils.strip(token.value)
            prefix += expr_prefix

            if self.check(tokens.ExpressionBlockEndToken):
                self.advance()
            else:
                return None
            suffix = self.consume_trivia() if parse_suffix else ""
            if self.log_debug:
                logger.debug("%sSuccessfully parsed value.", self.depth * DELIMITER)
            return tree.ExpressionSyntaxToken(
                value, token.line_number, prefix, suffix, expr_suffix
            )
        else:
            return None

    @backtrack_if_none
    def parse_list(self) -> Optional[tree.ListNode]:
        """Returns a node that represents a LookML list.

        Grammar: ``list`` ← ``key "[" csv? "]"``

        Returns:
            A node with the parsed list or None if the grammar doesn't match

        """
        if self.log_debug:
            grammar = "[list] = key '[' csv? ']'"
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        key = self.parse_key()
        if key is None:
            return key

        prefix = self.consume_trivia()
        if self.check(tokens.ListStartToken):
            self.advance()
            left_bracket = tree.LeftBracket(prefix=prefix)
        else:
            return None

        csv = self.parse_csv()
        csv = csv or CommaSeparatedValues()

        if self.check(tokens.ListEndToken, skip_trivia=True):
            prefix = self.consume_trivia()
            self.advance()
            suffix = self.consume_trivia()
            right_bracket = tree.RightBracket(prefix=prefix, suffix=suffix)
            node = tree.ListNode(
                type=key[0],
                colon=key[1],
                left_bracket=left_bracket,
                items=csv.values,
                right_bracket=right_bracket,
                leading_comma=csv.leading_comma,
                trailing_comma=csv.trailing_comma,
            )
            if self.log_debug:
                logger.debug("%sSuccessfully parsed a list.", self.depth * DELIMITER)
            return node
        else:
            return None

    @backtrack_if_none
    def parse_csv(self) -> Optional[CommaSeparatedValues]:
        """Returns a CSV object that represents comma-separated LookML values.

        Returns:
            A container with the parsed values or None if the grammar doesn't match

        Grammar:
            ``csv`` ←
            ``","? (literal / quoted_literal) ("," (literal / quoted_literal))* ","?``

        """
        if self.log_debug:
            grammar = (
                '[csv] = ","? (literal / quoted_literal) '
                '("," (literal / quoted_literal))* ","?'
            )
            logger.debug("%sTry to parse %s", self.depth * DELIMITER, grammar)

        # Set a flag to ensure that all items are of the same type (pair or literal)
        pair_mode: bool = False
        csv = CommaSeparatedValues()
        csv.leading_comma = self.parse_comma()

        # Parse the first value to set the list's type
        pair = self.parse_pair()
        if pair is not None:
            csv.append(pair)
            pair_mode = True
        elif self.check(
            tokens.LiteralToken, tokens.QuotedLiteralToken, skip_trivia=True
        ):
            value = self.parse_value(parse_prefix=True, parse_suffix=True)
            csv.append(value)
        else:
            return None

        # Parse to the closing bracket of the list
        while not self.check(tokens.ListEndToken, skip_trivia=True):
            if self.check(tokens.CommaToken):
                index = self.index
                self.advance()
                if self.check(tokens.ListEndToken, skip_trivia=True):
                    self.jump_to_index(index)  # Return to the comma so we can parse it
                    csv.trailing_comma = self.parse_comma()
                    break
            else:
                return None

            if pair_mode:
                pair = self.parse_pair()
                if pair is None:
                    return None
                else:
                    csv.append(pair)
            else:
                if self.check(
                    tokens.LiteralToken, tokens.QuotedLiteralToken, skip_trivia=True
                ):
                    value = self.parse_value(parse_prefix=True, parse_suffix=True)
                    csv.append(value)
                elif self.check(tokens.ListEndToken, skip_trivia=True):
                    break
                else:
                    return None

        if self.log_debug:
            logger.debug(
                "%sSuccessfully parsed comma-separated values.", self.depth * DELIMITER
            )
        return csv

    @backtrack_if_none
    def parse_comma(self) -> Optional[tree.Comma]:
        prefix = self.consume_trivia()
        if self.check(tokens.CommaToken):
            self.advance()
            if not self.check(tokens.ListEndToken, skip_trivia=True):
                suffix = self.consume_trivia()
            else:
                suffix = ""
            return tree.Comma(prefix=prefix, suffix=suffix)
        else:
            return None



import logging
from dataclasses import replace
from typing import Union, overload

from lkml.tree import (
    BlockNode,
    ContainerNode,
    DocumentNode,
    ListNode,
    PairNode,
    SyntaxNode,
    SyntaxToken,
    Visitor,
)

logger = logging.getLogger(__name__)


class BasicVisitor(Visitor):
    """Visitor class that calls the ``_visit`` method for every node type.

    This class doesn't actually do anything when visiting a tree other than traverse
    the nodes. It's meant to be used as a base class for building more useful and
    complex visitors. For example, override any of the ``visit_`` methods for node-type
    specific behavior.

    """

    def _visit(self, node: Union[SyntaxNode, SyntaxToken]):
        """For each node, visit its children."""
        if isinstance(node, SyntaxToken):
            return
        elif node.children:
            for child in node.children:
                child.accept(self)

    def visit(self, document: DocumentNode):
        return self._visit(document)

    def visit_container(self, node: ContainerNode):
        return self._visit(node)

    def visit_block(self, node: BlockNode):
        return self._visit(node)

    def visit_list(self, node: ListNode):
        return self._visit(node)

    def visit_pair(self, node: PairNode):
        return self._visit(node)

    def visit_token(self, token: SyntaxToken):
        return self._visit(token)


class LookMlVisitor(BasicVisitor):
    """Converts a parse tree into a string by casting every node."""

    @staticmethod
    def _visit(node: Union[SyntaxNode, SyntaxToken]) -> str:
        return str(node)


class BasicTransformer(Visitor):
    """Visitor class that returns a new tree, modifying the tree as needed."""

    @overload
    def _visit_items(self, node: ContainerNode) -> ContainerNode:
        ...

    @overload
    def _visit_items(self, node: ListNode) -> ListNode:
        ...

    def _visit_items(self, node):
        """Visit a node whose children are held in the ``items`` attribute."""
        if node.children:
            new_children = tuple(child.accept(self) for child in node.children)
            return replace(node, items=new_children)
        else:
            return node

    @overload
    def _visit_container(self, node: BlockNode) -> BlockNode:
        ...

    @overload
    def _visit_container(self, node: DocumentNode) -> DocumentNode:
        ...

    def _visit_container(self, node):
        """Visit a node whose only child is the ``container`` attribute."""
        if node.container:
            new_child = node.container.accept(self)
            return replace(node, container=new_child)
        else:
            return node

    def visit(self, node: DocumentNode) -> DocumentNode:
        return self._visit_container(node)

    def visit_container(self, node: ContainerNode) -> ContainerNode:
        return self._visit_items(node)

    def visit_list(self, node: ListNode) -> ListNode:
        return self._visit_items(node)

    def visit_block(self, node: BlockNode) -> BlockNode:
        return self._visit_container(node)

    def visit_pair(self, node: PairNode) -> PairNode:
        return node

    def visit_token(self, token: SyntaxToken) -> SyntaxToken:
        return token



from typing import Tuple


def strip(s: str) -> Tuple[str, str, str]:
    """Modified str.strip() which returns the stripped whitespace."""
    start = 0
    end = -1
    nchars = len(s)

    while start < nchars and s[start].isspace():
        start += 1
    if start != nchars:  # Check for case where only whitespace
        while s[end].isspace():
            end -= 1

    lspace = "" if start == 0 else s[:start]
    rspace = "" if end == -1 else s[end + 1 :]

    return lspace, s[start : end + 1 or None], rspace



from pathlib import Path

from setuptools import find_packages, setup

__version__ = "1.3.0b1"

here = Path(__file__).parent.resolve()

# Get the long description from the README file
with (here / "README.md").open(encoding="utf-8") as file:
    long_description = file.read()

setup(
    name="lkml",
    version=__version__,
    description="A speedy LookML parser implemented in pure Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/joshtemple/lkml",
    download_url="https://github.com/joshtemple/lkml/tarball/" + __version__,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
    ],
    keywords="lookml looker parser",
    license="MIT",
    entry_points={"console_scripts": ["lkml = lkml.__init__:cli"]},
    packages=find_packages(exclude=["docs", "tests*", "scripts"]),
    package_data={"lkml": ["py.typed"]},
    include_package_data=True,
    author="Josh Temple",
    tests_require=["black", "flake8", "isort", "mypy", "pytest", "pytest-cov"],
    author_email="",
    zip_safe=False,
)
